bl_info = {
    "name": "Export Curve to GameSynth",
    "author": "Tsugi <www.tsugi-studio.com>",
    "version": (1, 0),
    "blender": (2, 83, 0),
    "location": "Graph Editor > F-Curve context menu -> Export to GameSynth",
    "description": "Exports an F-Curve as a GameSynth automation curve.",
    "doc_url": "http://www.tsugi-studio.com/Software/GameSynth/API/en/examples.html#blender",
    "category": "Import-Export",
}

import bpy
import os
import re
import socket
import sys

from bpy import context
from bpy.types import Operator
from bpy_extras.object_utils import AddObjectHelper, object_data_add


class GRAPH_OT_ExportCurve(bpy.types.Operator):
    """Export an F-Curve as an automation curve."""   
    bl_idname = "export.gamesynth_curve"
    bl_label = "Export to GameSynth"
    bl_options = {'REGISTER'}

    # GS API communication settings
    port = 28542
    delimiter = "\r"
    
    # connect to the GS API
    def connect(self):
        try:
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.connect(("localhost", self.port))
            self.report({'OPERATOR'}, "GSAPI: Sucessfully connected to GameSynth.")   
            return s
        except ConnectionRefusedError:
            self.report({'ERROR'}, "GSAPI: Could not connect to GameSynth, please start GameSynth first.")   
            return ""

    # disonnect from the GS API
    def disconnect(self, s):
        s.close()
        self.report({'OPERATOR'}, "GSAPI: Disconnected from GameSynth.")

    # calculate duration of an F-Curve
    def calculate_duration(self, fcurve):
        start_frame = fcurve.range()[0]
        end_frame = fcurve.range()[1]
        frame_rate = bpy.context.scene.render.fps / bpy.context.scene.render.fps_base
        start_time = start_frame / frame_rate
        end_time = end_frame / frame_rate
        duration = end_time - start_time 
        return duration

    # make sure that a string has no unwanted characters
    def validate_name(self, name):
        pattern = r'[^a-zA-Z0-9_ ]'
        name = re.sub(pattern, '_', name)
        name = re.sub(r'_+', '_', name)
        return name
        
    def execute(self, context):
            
        # initialize connection        
        s = self.connect()
        if s == "":        
            return {'FINISHED'}
        
        # assuming there is a selected object with animation data, get the first F-Curve selected
        obj = bpy.context.active_object        
        if obj.animation_data and obj.animation_data.action:
            duration = 0
            for fcurve in obj.animation_data.action.fcurves:
                if fcurve.select:
                    self.report({'INFO'}, f"Selected F-Curve for property '{fcurve.data_path}' on channel {fcurve.array_index}")
                    
                    # get the automation curve points by evaluating the F-Curve for every frame in the range
                    points = []
                    for i in range(int(fcurve.range()[0]),int(fcurve.range()[1])):
                        context.scene.frame_set(i)
                        x, y = i, fcurve.evaluate(i)
                        points.append((x, y))
                    
                    # get its duration and name   
                    duration = self.calculate_duration(fcurve)
                    fcurve_name = fcurve.data_path + "_" + str(fcurve.array_index);
                    fcurve_name = self.validate_name(fcurve_name)
                    break
        else:
            self.report({'WARNING'},"The active object has no animation data.")
            return {'FINISHED'}
    
        if duration == 0:
            self.report({'WARNING'},"No F-Curve is currently selected.") 
            return {'FINISHED'}   
    
        # get the name of the current GameSynth patch
        s.send(("get_patchname" + self.delimiter).encode('utf-8'))
        patch_name = s.recv(self.port).decode('utf-8')
        patch_name = patch_name.rstrip(self.delimiter)
        if "ERROR: no patch loaded" in patch_name:
            self.report({'ERROR'}, "GSAPI: ERROR: No patch loaded.")
            self.disconnect(s)
            return {'FINISHED'}
        self.report({'INFO'}, "GSAPI: Current patch is " + patch_name + ".")  
        
        # get the list of possible target curves  
        s.send(("get_curvenames" + self.delimiter).encode('utf-8'))
        curve_names = s.recv(self.port).decode('utf-8')  
        curve_names = curve_names.rstrip(self.delimiter)
        if curve_names == "No data found":
            self.report({'INFO'}, "GSAPI: There are no automation curves in the patch.")
            curve_names = fcurve_name 
        else:
            self.report({'INFO'}, f"GSAPI: The current curves are {curve_names}.")  
            curve_names = fcurve_name + "," + curve_names 
          
        # display settings window
        header_label = "{LABEL, \"Curve Export Settings\", HEADER, CENTER}"
        automation_enum = "{ENUM, \"Target Automation Curve\", COMBO, \"" + curve_names + "\"," + fcurve_name + "}"
        processing_enum = "{ENUM, \"Processing\", COMBO, \"None, Clip, Normalize\", Normalize}"
        duration_float = "{NUMBER, \"Duration\", FLOAT, \"s\", 0.01, 300," + str(duration) + ", 2}"
        loop_bool = "{BOOL, \"Looping curve\", FALSE}"
        window_parameters = f"window_parameters {header_label} {automation_enum} {processing_enum} {duration_float} {loop_bool}"   
        s.send((window_parameters + self.delimiter).encode('utf-8'))

        # get settings
        result = s.recv(self.port).decode('utf-8')      
        result = result.rstrip(self.delimiter)
        parameters = result.split(',')      
        target, processing, duration, loop = parameters

        # change parameters to the right format
        bool_mapping = {"FALSE": 0, "TRUE": 1}
        loop = bool_mapping[loop]
        duration = float(duration)
        
        # process the points 
        min_x = min(point[0] for point in points)
        max_x = max(point[0] for point in points)
        points = [( (x - min_x) / (max_x - min_x), y) for x, y in points]        
        if processing == "Clip":
            points = [(x, min(max(y, 0), 1)) for x, y in points]
        else:
            if processing == "Normalize":
                min_y = min(point[1] for point in points)
                max_y = max(point[1] for point in points)
                if min_y != max_y:
                    points = [(x, (y - min_y) / (max_y - min_y)) for x, y in points]
                else:
                    points = [(x, max(0, min(y, 1))) for x, y in points]

        # send the curve
        set_curvevalue = f"set_curvevalue BY_NAME \"{target}\" {points} {duration} {loop}"
        set_curvevalue = set_curvevalue.replace("[","\"").replace("]","\"")
        self.report({'OPERATOR'}, f"set_curvevalue BY_NAME \"{target}\"")
        s.send((set_curvevalue + self.delimiter).encode('utf-8'))
        result = s.recv(self.port).decode('utf-8')
        result = result.rstrip(self.delimiter)
        if result != "OK":
            self.report({'ERROR'}, "GSAPI: " + result)   
    
        self.disconnect(s)
              
        return {'FINISHED'}
    

def menu_func(self, context):
    self.layout.operator(GRAPH_OT_ExportCurve.bl_idname)


def register():
    bpy.utils.register_class(GRAPH_OT_ExportCurve)
    bpy.types.GRAPH_MT_context_menu.append(menu_func)

def unregister():
    bpy.utils.unregister_class(GRAPH_OT_ExportCurve)
    bpy.types.GRAPH_MT_context_menu.remove(menu_func)

if __name__ == "__main__":
    register()
