bl_info = {
    "name": "Render GameSynth Patch",
    "author": "Tsugi <www.tsugi-studio.com>",
    "version": (1, 0),
    "blender": (2, 83, 0),
    "location": "Video Sequencer > Sequencer > Add > GameSynth Patch",
    "description": "Renders a sound variation at each selected marker's position in the Sequencer",
    "doc_url": "http://www.tsugi-studio.com/Software/GameSynth/API/en/examples.html#blender",
    "category": "Sequencer",
}

import bpy
import os
import socket
import sys

from bpy import context
from bpy.types import Operator
from bpy_extras.object_utils import AddObjectHelper, object_data_add


class SEQUENCER_OT_RenderPatch(bpy.types.Operator):
    """Render a sound variation on each selected marker, or on cursor."""   
    bl_idname = "sequencer.gamesynth_patch"
    bl_label = "GameSynth Patch"
    bl_options = {'REGISTER', 'UNDO'}

    # GS API communication settings
    port = 28542
    delimiter = "\r"
    
    # connect to the GS API
    def connect(self):
        try:
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.connect(("localhost", self.port))
            self.report({'OPERATOR'}, "GSAPI: Sucessfully connected to GameSynth.")   
            return s
        except ConnectionRefusedError:
            self.report({'ERROR'}, "GSAPI: Could not connect to GameSynth, please start GameSynth first.")   
            return ""

    # disonnect from the GS API
    def disconnect(self, s):
        s.close()
        self.report({'OPERATOR'}, "GSAPI: Disconnected from GameSynth.")

    # render the current patch
    def render_sound(self, s, file_path, bit_depth, outputs, duration):
        render_patch = f"render_patch \"{file_path}\" {bit_depth} {outputs} {duration}"
        self.report({'OPERATOR'}, "GSAPI: " + render_patch)
        render_patch = s.send((render_patch + self.delimiter).encode('utf-8'))
        result = s.recv(self.port).decode('utf-8')
                             
    # add a sound to the sequence
    def add_sound(self, scene, file_path,channel,frame):
        name = os.path.splitext(os.path.basename(file_path))[0]
        sound_strip = scene.sequence_editor.sequences.new_sound(name, file_path, channel, frame)
        sound_strip.show_waveform = True


    def execute(self, context):
            
        # initialize connection
        s = self.connect()
        if s == "":        
            return {'FINISHED'}

        # get the name of the current GameSynth patch
        s.send(("get_patchname" + self.delimiter).encode('utf-8'))
        patch_name = s.recv(self.port).decode('utf-8')
        patch_name = patch_name.rstrip(self.delimiter)
        if "ERROR: no patch loaded" in patch_name:
            self.report({'ERROR'}, "GSAPI: ERROR: No patch loaded.")
            self.disconnect(s)
            return {'FINISHED'}
        self.report({'INFO'}, "GSAPI: Current patch is " + patch_name + ".")  
          
        # check if the patch is infinite
        s.send("is_infinite".encode('utf-8'))
        is_infinite = s.recv(self.port).decode('utf-8')
        is_infinite = int(str(is_infinite).replace(self.delimiter, ""))

        # get the rendering parameters
        folder = bpy.path.abspath("//audio");
        window_label = "{LABEL, \"Select the Rendering Settings\", HEADER, CENTER}"
        folder_string = '{{STRING, "Rendering folder", FOLDER, "{0}"}}'.format(folder)
        bitdepth_enum = "{ENUM, \"Bit depth\", COMBO, \"16-bit,24-bit\", 16-bit}"
        outputs_enum = "{ENUM, \"Audio outputs\", COMBO, \"Mono,Stereo\", Mono}"
        duration_float = "{NUMBER, \"Duration\", FLOAT, \"s\", 0.1, 180, 10, 2}"
        channel_int = "{NUMBER, \"Destination channel\", INTEGER, \"\", 1, 128, 1, 0}"

        if is_infinite == 1:  
            window_parameters = f"window_parameters {window_label} {folder_string} {bitdepth_enum} {outputs_enum} {duration_float} {channel_int}"
        else:
            window_parameters = f"window_parameters {window_label} {folder_string} {bitdepth_enum} {outputs_enum} {channel_int}"   
        s.settimeout(None) # should be by default, but having some tools like WinSCP open in the background can override it
        s.send((window_parameters + self.delimiter).encode('utf-8'))

        result = s.recv(self.port).decode('utf-8')      
        result = result.rstrip(self.delimiter)
        parameters = result.split(',')      
        duration = 0
        if is_infinite == 1:
            folder, bit_depth, outputs, duration, channel = parameters
        else:      
            folder, bit_depth, outputs, channel = parameters

        # change parameters to the right format
        outputs_mapping = {"Mono": 1, "Stereo": 2}
        bit_depth_mapping = {"16-bit": 16, "24-bit": 24}
        outputs = outputs_mapping[outputs]
        bit_depth = bit_depth_mapping[bit_depth]
        duration = float(duration)
        channel = int(channel)

        # stop the script if the patch is infinite and the duration is 0
        if (is_infinite == 1 and duration <= 0):
            self.report({'ERROR'}, "GSAPI: ERROR: The patch is infinite, please set a duration higher than 0.")
            self.disconnect(s)
            return {'FINISHED'}
            
        scene = context.scene 

        if not scene.sequence_editor:
            scene.sequence_editor_create()

        if not os.path.exists(folder):
            try:
                os.makedirs(folder)
            except:
                self.report({'ERROR'},"GSAPI: ERROR: Could not create the folder " + folder)
                self.disconnect(s)
                return {'FINISHED'}

        # if there are selected markers we render a sound variations on each of them
        index = 0
        base_name = folder + "\\" + patch_name
        for marker in scene.timeline_markers:
             if marker.select:    
                index+=1;    
                file_path = base_name + "_" + str(index) + ".wav"
                self.render_sound(s, file_path, bit_depth, outputs, duration)
                self.add_sound(scene, file_path, channel, marker.frame)

        # otherwise we render a single sound at the cursor position
        if index == 0:
            current_frame = bpy.context.scene.frame_current    
            file_path = base_name + ".wav"
            self.render_sound(s, file_path, bit_depth, outputs, duration)
            self.add_sound(scene, file_path, channel, current_frame)

        self.disconnect(s)
              
        return {'FINISHED'}
    

def menu_func(self, context):
    self.layout.operator(SEQUENCER_OT_RenderPatch.bl_idname)


def register():
    bpy.utils.register_class(SEQUENCER_OT_RenderPatch)
    bpy.types.SEQUENCER_MT_add.append(menu_func)

def unregister():
    bpy.utils.unregister_class(SEQUENCER_OT_RenderPatch)
    bpy.types.SEQUENCER_MT_add.remove(menu_func)


if __name__ == "__main__":
    register()
